<?php
namespace Undefined\TranslateLocallang\Controller;
/***************************************************************
 *  Copyright notice
 *
 *  (c) 2016 Raphael Graf <r@undefined.ch>
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Extbase\Utility\LocalizationUtility;
use TYPO3\CMS\Core\Messaging\AbstractMessage;
use TYPO3\CMS\Core\Utility\ExtensionManagementUtility;
use Undefined\TranslateLocallang\Utility\TranslateUtility;

class ModuleController extends \TYPO3\CMS\Extbase\Mvc\Controller\ActionController
{
	/**
	 * @var array
	 */
	private $conf = [];

	/**
	 * @return void
	 */
	public function initializeAction() {
		$conf = unserialize($GLOBALS['TYPO3_CONF_VARS']['EXT']['extConf']['translate_locallang']);
		$this->conf['defaultLangKey'] = (trim($conf['defaultLangKey'])) ? trim($conf['defaultLangKey']) : 'en';
		$langKeys = GeneralUtility::trimExplode(',', $conf['langKeys']);
		$this->conf['langKeys'] = array_merge(['default' => $this->conf['defaultLangKey'] . ' (default)'], array_combine($langKeys, $langKeys));
		$allowedExts = GeneralUtility::trimExplode(',', $conf['allowedExts']);
		$this->conf['extensions'] = TranslateUtility::getExtList($allowedExts);
		$this->conf['modifyKeys'] = (bool)$conf['modifyKeys'] || $GLOBALS['BE_USER']->user['admin'];
		$this->conf['modifyDefaultLang'] = (bool)$conf['modifyDefaultLang'] || $GLOBALS['BE_USER']->user['admin'] || $this->conf['modifyKeys'];
		$this->conf['useL10n'] = (bool)$conf['useL10n'];
		$this->conf['debug'] = (bool)$conf['debug'];
		$this->conf['sysLog'] = (bool)$conf['sysLog'];
		$this->conf['langKeysAllowed'] = $this->conf['langKeys'];
		if (!$this->conf['modifyDefaultLang'])
			unset($this->conf['langKeysAllowed']['default']);
	}

	/**
	 * @param string extension
	 * @param string file
	 * @param array langKeys
	 * @return void
	 */
	public function listAction($extension = '', $file = '', $langKeys = ['default'])
	{
		$moduledata = TranslateUtility::getModuleData();
		if ($moduledata) {
			if (!$extension && $extension !== '0') {
				//restore from moduledata
				if ($moduledata['extension'] && isset($this->conf['extensions'][$moduledata['extension']]))
					$extension = $moduledata['extension'];
				if ($moduledata['file'])
					$file = $moduledata['file'];
				if ($moduledata['langKeys'])
					$langKeys = $moduledata['langKeys'];
			}
			if ($moduledata['sessid'] !== $GLOBALS['BE_USER']->id && $moduledata['extension'] === $extension) {
				$timediff = time()- $moduledata['time'];
				if ($timediff < 600) {
					$timestr = ($timediff < 60) ? $timediff . ' seconds' : (int)(($timediff + 30) / 60) . ' minutes';
					$this->addFlashMessage('Someone else started editing this extension ' . $timestr . ' ago.', 'Warning', AbstractMessage::WARNING);
				}
			}
		}

		if (!in_array('default', $langKeys))
			array_unshift($langKeys, 'default'); //default is always shown
		
		if (!$extension)
			$file = '';

		if ($extension) {
			if (!isset($this->conf['extensions'][$extension]))
				return __FUNCTION__ . ': extension not allowed: ' . $extension;  //ERR
			$l = next($this->conf['langKeys']);
			$l10ndir = 'l10n/' . $l . '/' . $extension;
			if (!$this->conf['useL10n'] && is_dir(PATH_typo3conf . $l10ndir))
				$this->addFlashMessage(
					'typo3conf/' . $l10ndir . ' directory exists. (You are currently editing the files in typo3conf/ext).', 'Notice', AbstractMessage::NOTICE
				);
			$files = TranslateUtility::getFileList($extension);
			array_unshift($files, LocalizationUtility::translate('choose_file', 'translate_locallang'));
			if ($file && !isset($files[$file]))
				$file = '';
			if ($file) {
				$xliffService = GeneralUtility::makeInstance('Undefined\TranslateLocallang\Service\XliffService');
				$xliffService->init($extension, $file, $this->conf['defaultLangKey'], $this->conf['useL10n']);
				if ($xliffService->isOldXMLFormat())
					$this->addFlashMessage('This file uses the old XML format. It will be saved with a "_converted.xlf" suffix.', 'Warning', AbstractMessage::WARNING);

				foreach($langKeys as $langKey) {
					if (!$xliffService->loadLang($langKey)) {
						if ($langKey === 'default')
							$file = ''; // do not error out (can happen when $extension was changed)
						else
							$xliffService->addLang($langKey);
					}
				}
				$labels = &$xliffService->getData();
				if (empty($labels))
					$this->addFlashMessage('No labels found.', 'Warning', AbstractMessage::WARNING);
				$max_input_vars = (int)ini_get('max_input_vars');
				$fieldcount = count($labels) * (count($langKeys) + 1) + count($langKeys) + 4;
				if ($fieldcount > $max_input_vars)
					$this->addFlashMessage('Too many labels, max_input_vars too small. Do not press save!', 'Warning', AbstractMessage::WARNING);
			}
		}

		$this->view->assignMultiple([
			'extensions' => array_merge(['0' => LocalizationUtility::translate('choose_extension', 'translate_locallang')], $this->conf['extensions']),
			'extension' => $extension,
			'files' => $files,
			'file' => $file,
			'langKeys' => $langKeys,
			'labels' => $labels,
			'conf' => $this->conf,
			'showCsh' => ExtensionManagementUtility::isLoaded('cshmanual'),
			'typo3Ver' => TYPO3_version[0],
		]);

		TranslateUtility::setModuleData(
			[
				'extension' => $extension,
				'file' => $file,
				'langKeys' => $langKeys,
				'time' => time(),
				'sessid' => $GLOBALS['BE_USER']->id,
			]
		);
	}

	/**
	 * @param array keys
	 * @param array labels
	 * @param string extension
	 * @param string file
	 * @param array langKeys
	 * @return void
	 */
	public function saveAction($keys = NULL, $labels = NULL, $extension = '', $file = '', $langKeys = ['default'])
	{
		if (!$keys) //this can happen via backend bookmark
			$this->forward('list', NULL, NULL, ['extension' => $extension, 'file' => $file, 'langKeys' => $langKeys]);

		if (!isset($this->conf['extensions'][$extension]))
			return __FUNCTION__ . ': extension not allowed: ' . $extension; //ERR
		$files = TranslateUtility::getFileList($extension);
		if (!isset($files[$file]))
			return __FUNCTION__ . ': file not allowed: ' . $file; //ERR

		foreach($langKeys as $key => $langKey) {
			if (!isset($this->conf['langKeysAllowed'][$langKey]))
				unset($langKeys[$key]);
		}

		//remove dummy keys
		foreach($keys as $key => $keyvalue) {
			if ($key !== $keyvalue) {
				if (strpos($key, '_newkey') === 0 && $keyvalue === '') {
					unset($labels[$key]);
					unset($keys[$key]);
				}
			}
		}

		$xliffService = GeneralUtility::makeInstance('Undefined\TranslateLocallang\Service\XliffService');
		$xliffService->init($extension, $file, $this->conf['defaultLangKey'], $this->conf['useL10n']);
		$xliffService->setData($labels);

		//handle keychanges
		$keychanges = [];
		foreach($keys as $key => $keyvalue) {
			if ($key !== $keyvalue) {
				if (!$this->conf['modifyKeys'])
					return __FUNCTION__ . ': not allowed to modify keys'; //ERR
				$keychanges[$key] = $keyvalue;
				$this->log('Changed key: ' . $extension . '|' . $file . ' ' . $key . '->' . $keyvalue, 0);
			}
		}

		$savelangs = $langKeys; //languages that need to be saved
		if (count($keychanges)) {
			//load all languages and change keys
			foreach($this->conf['langKeysAllowed'] as $langKey => $dummy) {
				if (!$xliffService->isLoaded($langKey)) {
					if (!$xliffService->loadLang($langKey))
						continue;
					$savelangs[] = $langKey;
				}
			}
			foreach($keychanges as $key => $keyvalue)
				$xliffService->changeKey($key, $keyvalue);
		}

		//save languages
		foreach($savelangs as $langKey) {
			if (!$xliffService->saveLang($langKey))
				$this->log('Write failed: ' . $xliffService->getFilename($langKey), 2);
		}

		$this->log('Updated ' . implode(', ', $savelangs), 0);
		$this->forward('list', NULL, NULL, ['extension' => $extension, 'file' => $file, 'langKeys' => $langKeys]);
	}

	/**
	 * @param string extension
	 * @param string file
	 * @param boolean defaultOnly
	 * @param string format
	 * @return void
	 */
	public function exportCsvAction($extension = '', $file = '', $defaultOnly = FALSE, $format = 'CSV')
	{
		if (!isset($this->conf['extensions'][$extension]))
			return __FUNCTION__ . ': extension not allowed: ' . $extension; //ERR
		$files = TranslateUtility::getFileList($extension);
		if (!isset($files[$file]))
			return __FUNCTION__ . ': file not allowed: ' . $file; //ERR

		$xliffService = GeneralUtility::makeInstance('Undefined\TranslateLocallang\Service\XliffService');
		$xliffService->init($extension, $file, $this->conf['defaultLangKey'], $this->conf['useL10n']);

		$langKeys = ($defaultOnly) ? ['default' => $this->conf['defaultLangKey'] . ' (default)'] : $this->conf['langKeys'];
		foreach($langKeys as $langKey => $dummy) {
			if (!$xliffService->loadLang($langKey)) {
				continue;
			}
		}

		$data = &$xliffService->getData();
		$hrow = ['key'];
		foreach($this->conf['langKeys'] as $langKey => $langKeyValue) {
			$hrow[] = $langKeyValue; //header row
		}

		if ($format == 'CSV') {
			//output CSV
			$fileName = $extension . '-' . $file . '.csv';
			header('Content-Type: text/x-csv');
			header('Content-Disposition: attachment; filename="' . $fileName . '"');
			header('Pragma: no-cache');
			print(pack('CCC', 239, 187, 191)); //BOM
			print(implode(';', $hrow) . "\n");
			foreach($data as $key => $labels) {
				$row = [$key];
				foreach($labels as $langKey => $dummy) {
					if (isset($labels[$langKey]))
						$row[] = '"' . $labels[$langKey] . '"'; //XXX addslashes or similar?
					else
						$row[] = '""';
				}
				print(implode(';', $row) . "\n");
			}
			return '';
		} else {
			//output FODS
			$fileName = $extension . '-' . $file . '.fods';
			header('Content-Type: application/vnd.oasis.opendocument.spreadsheet');
			header('Content-Disposition: attachment; filename="' . $fileName . '"');
			header('Pragma: no-cache');
			$fodsview = $this->objectManager->get('TYPO3\\CMS\\Fluid\\View\\StandaloneView');
			$fodsview->setTemplatePathAndFilename(PATH_typo3conf . 'ext/translate_locallang/Resources/Private/Templates/Export/Fods.html');

			$fodsview->assignMultiple([
				'header' => $hrow,
				'labels' => $data
			]);
			return $fodsview->render();
		}
	}

	/**
	 * @param string msg
	 * @param int error (0 = message, 1 = User Error, 2 = System Error, 3 = security notice)
	 * @return boolean
	 */
	protected function log($msg, $error = 0) {
		if ($this->conf['sysLog'] || $error)
			$GLOBALS['BE_USER']->simplelog($msg, 'translate_locallang', $error);
		if ($this->conf['debug'] || $error)
			$this->addFlashMessage($msg, ($error) ? 'Error' : 'Debug', ($error) ? AbstractMessage::ERROR : AbstractMessage::NOTICE);
	}
}
