<?php
namespace Undefined\TranslateLocallang\Service;
/***************************************************************
 *  Copyright notice
 *
 *  (c) 2016 Raphael Graf <r@undefined.ch>
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use TYPO3\CMS\Core\Utility\GeneralUtility;

class XliffService
{
	/**
	* @var array
	*/
	protected $data = [];

	/**
	* @var string
	*/
	protected $extension = NULL;

	/**
	* @var string
	*/
	protected $file = NULL;

	/**
	* @var string
	*/
	protected $useL10n = TRUE;

	/**
	* @var string
	*/
	protected $sourcelang = 'en';

	/**
	* read support for the old XML Format is inlined here.
	* this will not stay for very long..
	* @var boolean
	*/
	protected $oldXMLFormat = FALSE;

	/**
	 * @param string extension
	 * @param string file
	 * @param string sourcelang
	 * @param boolean useL10n
	 * @return void
	 */
	public function init($extension, $file, $sourcelang = 'en', $useL10n = TRUE) {
		$this->extension = $extension;
		$this->file = $file;
		$this->sourcelang = $sourcelang;
		$this->useL10n = $useL10n;
		$this->oldXMLFormat = strpos($file, '.xml') === strlen($file) - 4;
	}

	/**
	 * load language data from file
	 *
	 * @param string langKey
	 * @return boolean
	 */
	public function loadLang($langKey) {
		//Note: This does not use TYPO3\CMS\Core\Localization\Parser\XliffParser because of CDATA support
		//load default data first
		if (!$this->isLoaded('default')) {
			$fileref = $this->getXlfPath('default');
			if (!is_file($fileref))
				return FALSE;
			libxml_use_internal_errors(TRUE);
			$xml = simplexml_load_file($fileref, 'SimpleXMLElement');
			if ($xml === FALSE) {
				echo 'Error parsing file for default language. (Fill the form and save to overwrite the broken file) <br>libxml errors: ' . "<pre>\n";
				print_r(libxml_get_errors());
				echo '</pre>';
				$this->data = [];
				return TRUE;
			}

			$hasdata = ($this->oldXMLFormat) ? isset($xml->data) : isset($xml->file) && isset($xml->file->body);
			if (!$hasdata) {
				$this->data = [];
				return TRUE;
			}

			$children = [];
			if ($this->oldXMLFormat) {
				foreach($xml->data->children() as $languagekey) {
					$key = (string)$languagekey['index'];
					if ($key === 'default')
						$children = $languagekey->children();
				}
			} else {
				$children = $xml->file->body->children();
			}

			foreach ($children as $transunit) {
				if ($this->oldXMLFormat) {
					$key = (string)$transunit['index'];
					$this->data[$key]['default'] = (string)$transunit;
				}
				if (isset($transunit['id'])) {
					$key = (string)$transunit['id'];
					$this->data[$key] = [];
					$this->getLabel($transunit, 'default');
				}
			}
		}

		//load overlay data
		if ($langKey !== 'default') {
			$children = [];
			$fileref = $this->getXlfPath($langKey);
			if ($this->oldXMLFormat && !is_file($fileref))
				$fileref = $this->getXlfPath('default');

			if (!is_file($fileref))
				return FALSE;

			$xml = simplexml_load_file($fileref, 'SimpleXMLElement');
			if ($xml !== FALSE){
				if ($this->oldXMLFormat) {
					if (isset($xml->data)) {
						foreach($xml->data->children() as $languagekey) {
							$key = (string)$languagekey['index'];
							if ($key === $langKey)
								$children = $languagekey->children();
						}
					}
				} else {
					if (isset($xml->file) && isset($xml->file->body))
						$children = $xml->file->body->children();
				}

				foreach ($children as $transunit) {
					if ($this->oldXMLFormat) {
						$key = (string)$transunit['index'];
						$this->data[$key][$langKey] = (string)$transunit;
					}
					if (isset($transunit['id'])) {
						$this->getLabel($transunit, $langKey);
					}
				}
			}

			//set missing labels to ''
			foreach($this->data as $key => $value) {
				if (!isset($value[$langKey]))
					$this->data[$key][$langKey] = '';
			}
		}

		return TRUE;
	}

	/**
	 * add blank language data
	 *
	 * @param string langKey
	 * @return void
	 */
	public function addLang($langKey) {
		foreach($this->data as $key => $dummy) {
			$this->data[$key][$langKey] = '';
		}
	}

	/**
	 * @param string langKey
	 * @return boolean
	 */
	public function saveLang($langKey) {
		$labels = []; //matches the format used by TYPO3\CMS\Core\Localization\Parser\XliffParser
		foreach($this->data as $key => $dummy)
			if (isset($this->data[$key]['default']) && strpos($key, '_newkey') !== 0)
				$labels[$key] = [
					0 => [
						'source' => $this->encodeValue($this->data[$key]['default']),
						'target' => $this->encodeValue($this->data[$key][$langKey])
				]];

		$objectManager = GeneralUtility::makeInstance('TYPO3\\CMS\\Extbase\\Object\\ObjectManager');
		$xliffview = $objectManager->get('TYPO3\\CMS\\Fluid\\View\\StandaloneView');
		$xliffview->setTemplatePathAndFilename(PATH_typo3conf . 'ext/translate_locallang/Resources/Private/Templates/Xliff.html');

		date_default_timezone_set('UTC');
		$xliffview->assignMultiple(
			[
				'labels' => $labels,
				'sourcelang' => $this->sourcelang,
				'targetlang' => ($langKey === 'default') ? NULL: $langKey,
				'productname' => $this->extension,
				'date' => date('Y-m-d\TH:i:s\Z'), //date('c')
			]
		);

		$xliff = $xliffview->render();
		if ($xliff) {
			if ($this->oldXMLFormat)
				$this->file = str_replace('.xml', '_converted.xlf', $this->file); //XXX
			$oformat = $this->oldXMLFormat;
			$this->oldXMLFormat = FALSE;
			$fileref = $this->getXlfPath($langKey);
			$this->oldXMLFormat = $oformat;
			if ($this->useL10n && !is_dir(dirname($fileref))) {
				$folderCreateMask = (isset($GLOBALS['TYPO3_CONF_VARS']['SYS']['folderCreateMask'])) ? 
					octdec($GLOBALS['TYPO3_CONF_VARS']['SYS']['folderCreateMask']) : octdec('2775');
				if (!mkdir(dirname($fileref), $folderCreateMask, TRUE))
					return FALSE;
			}
			if ($handle = @fopen($fileref , 'w')) {
				if (fwrite($handle , $xliff) === FALSE)
					return FALSE;
				if (isset($GLOBALS['TYPO3_CONF_VARS']['SYS']['fileCreateMask']))
					@chmod($fileref, octdec($GLOBALS['TYPO3_CONF_VARS']['SYS']['fileCreateMask']));
				return TRUE;
			}
		}
		return FALSE;
	}

	/**
	 * @param \SimpleXMLElement $transunit
	 * @param string $langkey
	 * @return void
	 */
	 private function getLabel($transunit, $langKey = 'default') {
		$key = (string)$transunit['id'];
		$value = $str = '';
		if ($langKey === 'default' && isset($transunit->source)) {
			$value = (string)$transunit->source;
			$str = $transunit->source->asXML();
		} else if (isset($transunit->target)) {
			$value = (string)$transunit->target;
			$str = $transunit->target->asXML();
		}
		$s = 8;
		while(ctype_space($str[$s]))
			++$s;
		if (isset($str[$s]) && substr_compare($str, '<![CDATA[' , $s, 9) === 0)
			$value = '<![CDATA[' . $value . ']]>';
		$this->data[$key][$langKey] = $value;
	 }

	/**
	 * @param string str
	 * @return string
	 */
	private function encodeValue($str) {
		if (strpos(ltrim($str), '<![CDATA[') === 0)
			return ltrim($str);
		else
			return htmlspecialchars($str);
	}

	/**
	 * @return array
	 */
	public function &getData() {
		return $this->data; //return a reference to the array
	}

	/**
	 * @param array data
	 * @return void
	 */
	public function setData($data) {
		$this->data = $data;
	}

	/**
	 * @param string langKey
	 * @return string
	 */
	public function getFilename($langKey) {
		return $this->getXlfPath($langKey);
	}

	/**
	 * @param string oldKey
	 * @param string newKey
	 * @return void
	 */
	public function changeKey($oldKey, $newKey) {
		//replacing the array key is insanely expensive
		if(!isset($this->data[$oldKey]))
			return;

		if ($newKey == '') {
			unset($this->data[$oldKey]);
			return;
		}

		$keys = array_keys($this->data);
		$keys[array_search($oldKey, $keys)] = $newKey;

		$data = array_combine($keys, $this->data);
		$this->data = $data;
	}

	/**
	 * get path to XLF, no checking for file existence, arguments are checked by caller
	 *
	 * @param string extension
	 * @param string file
	 * @param string langKey
	 * @return string
	 */
	private function getXlfPath($langKey = 'default') {
		//get default path
		if ($langKey === 'default')
			return PATH_typo3conf . 'ext/' . $this->extension . '/Resources/Private/Language/' . $this->file;

		//get overlay path
		$file = $langKey . '.' . $this->file;
		if ($this->oldXMLFormat)
			$file = substr_replace($this->file, '.' . $langKey, strlen($this->file) - 4, 0);

		if ($this->useL10n)
			$path = PATH_typo3conf . 'l10n/' . $langKey . '/' . $this->extension . '/Resources/Private/Language/' . $file;
		else
			$path = PATH_typo3conf . 'ext/' . $this->extension . '/Resources/Private/Language/' . $file;
		return $path;
	}

	/**
	 * @param string langKey
	 * @return boolean
	 */
	public function isLoaded($langKey) {
		$elem = reset($this->data);
		return isset($elem[$langKey]);
	}

	/**
	 * @return boolean
	 */
	public function isOldXMLFormat() {
		return $this->oldXMLFormat;
	}
}
